/*  Copyright (C) 2001-2002 Vanja Hrustic

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA

    In addition, as a special exception, Vanja Hrustic
    gives permission to link the code of this program with
    the SAVI library, and distribute linked combinations including
    the two. You must obey the GNU General Public License in all
    respects for all of the code used other than SAVI. If you modify
    this file, you may extend this exception to your version of the
    file, but you are not obligated to do so. If you do not wish to
    do so, delete this exception statement from your version.

*/

#include "sophie.h"

/* $Id: sophie.c,v 1.1.1.1 2003/02/28 15:51:06 vanja Exp $ */

volatile int PROC_COUNT = 0;

/* Print the usage */
void usage(void)
{
	fprintf(stdout, "\n(Version: %s)\n\n", SOPHIE_VERSION);
	fprintf(stdout, "USAGE: %s [options]\n", program_name);
	fprintf(stdout, "Options:\n");
	fprintf(stdout, "  -c        Show compiled-in Sophos/SAVI engine configuration\n");
	fprintf(stdout, "  -D        Daemon mode (forks into background)\n");
	fprintf(stdout, "  -d        Debug mode (shows debugging output, does not fork into background)\n");
	fprintf(stdout, "  -f [path] Filename (to scan for viruses)\n");
	fprintf(stdout, "  -s [path] Directory (to scan for viruses)\n");
	fprintf(stdout, "  -h        Help (this screen)\n");
	fprintf(stdout, "  -v        SAVI version (engine and patterns)\n");
}

/*
   We don't want any zombies
   Hope this is the right way to do it :)
*/
void sig_chld(int sig)
{
	/* Suggested by philipp@corpex.de (Philipp Gaschtz) */
	int stat;
	pid_t cpid;
	int serrno = errno;

	/* while there are any children that have exited... */
	while((cpid = waitpid(-1, &stat, WNOHANG) > 0) && PROC_COUNT)
	{
		if (cpid == 0) { } /* no children here */

		if (cpid == -1)
		{
			sophie_print(1, "There has been a problem catching a child: %s", strerror(errno));
			break;
		}

		if (cpid > 0)
		{
			/* Should not happen, but... */
			if (PROC_COUNT > 0)
				PROC_COUNT--;

/*			sophie_print(1, "child finished/exited"); */
		}
	}

	errno = serrno;
}

/* Print PROC_COUNT */
void sig_usr1(int sig)
{
	sophie_print(0, "Current process count (PROC_COUNT): [%d]", PROC_COUNT);
}

/* Cleanup and exit */
void sig_exit(int sig)
{
	sophie_print(0, "SIGNAL [%d] caught - cleaning up and exiting.", sig);

	sophie_end();
	close(sock);
#ifdef SOPHIE_NET
	shutdown(tcp_sock, 2);
	close(tcp_sock);
#endif
	unlink(SOPHIE_SOCKET_NAME);
	unlink(SOPHIE_PIDFILE_NAME);
	exit(0);
}

/* Restart Sophie */
void sig_reload(int sig)
{
	sophie_end();
	close(sock);
#ifdef SOPHIE_NET
	shutdown(tcp_sock, 2);
	close(tcp_sock);
#endif
	unlink(SOPHIE_SOCKET_NAME);
	unlink(SOPHIE_PIDFILE_NAME);

	sophie_print(0, "Initializing    : Received -HUP signal [%d] - restarting", sig);

	execv(program_name, program_args);

	/* 2002/10/08 Anne Bennett: this should not be reached: */
	syslog(LOG_ERR, "Sophie cannot restart program '%s': %m", program_name);
	exit(1);
}

/* Catch SIGALRM delivered to parent process - still no clue why it happens */
void sig_palarm(int sig)
{
	syslog(LOG_ERR, "Sophie parent process has just received SIGALRM - resuming!\n");
}

void sophie_timeout(int sig)
{
	syslog(LOG_ERR, "Sophie child has timed-out (no data received in %d seconds) - process killed\n", SOPHIE_TIMEOUT);
	exit(-1);
}

/* Here we go... */
int main(int argc, char *argv[])
{
	/* Result returned by SAVI */
	int sophos_result = 0;
	
	/* Related to socket */
	int msgsock = -99;
	
	/* Related to other stuff */
	int c, pid, cpid;
	
	/* 'Read' buffer */
	char buf[MAXPATHLEN];
	
	/* Pathfile to scan */
	char scan[MAXPATHLEN];

	/* Path (directory) to scan */
	char dirscan[MAXPATHLEN];

	/* Response to send back */
	char sock_response[256];
	
	/* fileInfo used to check if the socket is really a socket :) (for IS_SOCK()) */
	struct stat fileInfo;
	
	/* socket structure... */
	struct sockaddr_un server;
	
	/* Used when changing the group of the socket */
	struct group *grpInfo;

	/* Used to check if the user exists */
	struct passwd *checkUser;

	/* Stupid signal handling code */
	int sig_ret;
	sigset_t set;
	sigset_t bset; /* blockset, for fork() */
	struct sigaction sa_chld;
	struct sigaction sa_usr1;

	/* child */
	struct sigaction sa_handle;
	struct stat reqstat;
	struct passwd *userInfo;

	/* temporary file descriptor to detach when daemonizing */
	int devnull;

	/* PID file */
	FILE *f;

	/* Scan type (local/network) */
	int scan_type = 0;

#ifdef SOPHIE_NET
	/* Network socket related stuff */
	int client_tcp_sock = -99;
	int reuse_addr = 1;
	struct protoent *proto;
	struct sockaddr_in tcp_sock_data;
/*   int childpid, clilen; */
#endif

	/* Copy the name of the program (as supplied on the command line) */
	program_name = argv[0];
	program_args = argv;

	sigemptyset(&set);
	sigprocmask(SIG_SETMASK, &set, NULL);

	signal(SIGALRM, SIG_IGN);
/*	signal(SIGCHLD, sig_chld); */
	signal(SIGHUP, sig_reload);
	signal(SIGINT, sig_exit);
	signal(SIGTERM, sig_exit);
	signal(SIGQUIT, sig_exit);
	signal(SIGSEGV, sig_exit);
	sophie_print(1, "Signal handlers set");

	/* SIGCHLD handler */
	sa_chld.sa_handler = sig_chld;
	sigfillset(&sa_chld.sa_mask);
	sa_chld.sa_flags = SA_RESTART;
	sig_ret = sigaction(SIGCHLD, &sa_chld, NULL);
	if (sig_ret < 0)
		sophie_print(0, "ERROR: Could not install SIGCHLD signal handler (sig_ret = [%d])", sig_ret);

	/* SIGUSR1 handler */
	sa_usr1.sa_handler = sig_usr1;
	sigfillset(&sa_usr1.sa_mask);
	sa_usr1.sa_flags = SA_RESTART;
	sig_ret = sigaction(SIGUSR1, &sa_usr1, NULL);
	if (sig_ret < 0)
		sophie_print(0, "ERROR: Could not install SIGUSR1 signal handler (sig_ret = [%d])", sig_ret);

	/* Clear the scan (path to the filename to scan */
	memset(scan, 0, sizeof(scan));
	memset(dirscan, 0, sizeof(dirscan));

	/* Parse cmdline arguments */
	while ((c = getopt(argc, argv, "cDdf:s:hv")) != EOF)
	{
		switch(c)
		{
			case 'c':
			sophie_init();
			sophie_show_settings();
			sophie_end();
			exit(0);
			break;

			case 'D':
			SOPHIE_DAEMON = 1;
			break;
			
			case 'd':
			SOPHIE_DEBUG = 1;
			break;

			case 'f':
			if (optarg != '\0')
				strncpy(scan, optarg, sizeof(scan)-1);
			break;
			
			case's':
			if (optarg != '\0')
				strncpy(dirscan, optarg, sizeof(dirscan)-1);
			break;

			case 'h':
			usage();
			exit(0);
			break;

			case 'v':
			SOPHIE_DAEMON = 0;
			SOPHIE_DEBUG = 0;
			sophie_init();
			sophie_version();
			sophie_end();
			exit(0);
			break;

			default:
			SOPHIE_DAEMON = 0;
			SOPHIE_DEBUG = 0;
			break;
		}
	}


	/* Make sure user really exists. Otherwise, Sophie will start, but childs will die if the user doesn't exist */
	if (!(checkUser = getpwnam(RUNAS_USER)))
	{
		sophie_print(0, "ERROR: RUNAS_USER (%s) does not exist - exiting", RUNAS_USER);

		if (SOPHIE_DAEMON == 1)
			fprintf(stderr, "ERROR: RUNAS_USER (%s) does not exist - exiting\n", RUNAS_USER);

		exit(1);
	}

	/* We will check if the SOPHIE_SOCKET_GROUP exists first */
	if (!getgrnam(SOPHIE_SOCKET_GROUP))
	{
		sophie_print(0, "ERROR: SOPHIE_SOCKET_GROUP group (%s) does not exist - exiting", SOPHIE_SOCKET_GROUP);

		if (SOPHIE_DAEMON == 1)
			fprintf(stderr, "ERROR: SOPHIE_SOCKET_GROUP group (%s) does not exist - exiting\n", SOPHIE_SOCKET_GROUP);

		exit(1);
	}

	/* If SOPHIE_DEBUG/SOPHIE_DAEMON are not set (0 or 1) = set to 0 */
	if (SOPHIE_DEBUG != 0 && SOPHIE_DEBUG != 1)
		SOPHIE_DEBUG = 0;
	
	if (SOPHIE_DAEMON != 0 && SOPHIE_DAEMON != 1)
		SOPHIE_DAEMON = 0;

	/* Initialize Sophos engine/data */
	sophie_init();

	/* Spit out the version number(s) */
	sophie_version();

	/* If file is specified on a command line, scan it and return result */
	/* When scan is finished, exit */
	if (scan[0] != '\0')
	{
		sophie_print(0, "Scanning file   : '%s' ", scan);
		sophos_result = sophie_scanfile(scan);
		sophie_end();
		sophie_print(0, "Cleanup         : Sophos cleaned up and terminated");
		exit(0);
	}
	
	/* Yes, I could combine the above with this, but for the time being I want to have them separated */
	if (dirscan[0] != '\0')
	{
		sophie_print(0, "Scanning dir    : '%s' ", dirscan);
		sophos_result = sophie_scandir(dirscan);
		sophie_end();
		sophie_print(0, "Cleanup         : Sophos cleaned up and terminated");
		exit(0);
	}


	/* Check if the socket already exists */
	if (stat(SOPHIE_SOCKET_NAME, &fileInfo) == 0)
	{
		if (S_ISSOCK(fileInfo.st_mode))
		{
			sophie_print(0, "ERROR: Socket '%s' already exists!", SOPHIE_SOCKET_NAME);
			sophie_print(0, "NOTE : Removing existing socket");
			if (unlink(SOPHIE_SOCKET_NAME) == -1)
			{
				sophie_print(0, "ERROR: Could not remove socket '%s'. Aborting!", SOPHIE_SOCKET_NAME);
				exit(1);
			}
		}
	}

	/* Create a socket */
	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
	{
		sophie_print(0, "ERROR: socket() failed in main()");
		exit(1);
	}

	/* Set the umask */
	umask(0007);
	
	/* Set the socket properties (family, name) */
	server.sun_family = AF_UNIX;
	strncpy(server.sun_path, SOPHIE_SOCKET_NAME, sizeof(server.sun_path)-1);
	sophie_print(1, "server.sun_path has a name: [%s]", SOPHIE_SOCKET_NAME);

	/* Bind a socket */
	if (bind(sock, (struct sockaddr *) &server, sizeof(struct sockaddr_un)) == -1)
	{
		sophie_print(0, "ERROR: bind() failed in main()");
		exit(1);
	}

	/* Change the owner/group of the socket */
	grpInfo = getgrnam(SOPHIE_SOCKET_GROUP);
	if (chown(SOPHIE_SOCKET_NAME, checkUser->pw_uid, grpInfo->gr_gid) == -1)
		sophie_print(0, "ERROR: failed to chown() the socket (user: %s, group %s)", RUNAS_USER, SOPHIE_SOCKET_GROUP);

	/* Start listening */
	if (listen(sock, SOPHIE_LISTEN_QUEUE) == -1)
	{
		sophie_print(0, "ERROR: listen() failed in main()");
		exit(1);
	}

	/* If SOPHIE_DAEMON is set, show some message */
	/* Argh - at this point, everything already goes to syslog (if SOPHIE_DAEMON is set) - will change this later :) */
	if (SOPHIE_DAEMON == 1)
	{
		if ((pid = fork()) > 0)
		{
			sophie_print(0, "%s placed in the background [PID: %d]", program_name, pid);
			exit(0);
		}
		else if (pid == -1)
		{
			perror("fork");
			sophie_print(0, "ERROR: Could not fork() - aborting");
			exit(1);
		}

		if (setsid() == -1)
		{
			perror("setsid");
			sophie_print(0, "ERROR: Could not set process group() - aborting");
			exit(1);
		}
		if ((devnull = open("/dev/null", O_RDWR, 0)) < 0)
		{
			perror("open");
			sophie_print(0, "ERROR: could not detach fds - aborting");
			exit(1);
		}

		fprintf(stderr, "%s placed in the background\n", program_name);
		(void)dup2(devnull, STDIN_FILENO);
		(void)dup2(devnull, STDOUT_FILENO);
		(void)dup2(devnull, STDERR_FILENO);
		if (devnull > 2)
			(void)close (devnull);
	}

	/* Create PID file */
	f = fopen(SOPHIE_PIDFILE_NAME, "wb");
	if (f)
	{
		fprintf(f, "%u\n", (u_int) getpid());
		fclose(f);
	}

#ifdef SOPHIE_NET
	if ((proto = getprotobyname("tcp")) == NULL)
	{
		sophie_print(0, "Could not get protocol number for TCP");
		exit(1);
	}

	if ((tcp_sock = socket(AF_INET, SOCK_STREAM, proto->p_proto)) < 0)
	{
		sophie_print(0, "Could not obtain a socket [tcp]");
		exit(1);
	}

	bzero((char *) &tcp_sock_data, sizeof(tcp_sock_data));
	tcp_sock_data.sin_family = AF_INET;
	tcp_sock_data.sin_addr.s_addr = htonl(INADDR_ANY); /* FIXME: configurable? */
	tcp_sock_data.sin_port = htons(SOPHIE_PORT);

	setsockopt(tcp_sock, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof(reuse_addr));

	if (bind(tcp_sock, (struct sockaddr *) &tcp_sock_data, sizeof(tcp_sock_data)) < 0)
	{
		perror("Could not bind local socket\n");
		exit(-1);
	}

	listen(tcp_sock, SOPHIE_NET_QUEUE);

	/* Make sockets non-blocking */
	fcntl(sock, F_SETFL, O_NONBLOCK);
	fcntl(tcp_sock, F_SETFL, O_NONBLOCK);
#endif

	/* Fun begins... */
	for (;;)
	{
LOOP:
		/* Accept the connection(s) */
		usleep(10);
#ifdef SOPHIE_NET
		if ( ((msgsock = accept(sock, 0, 0)) > 0) || ((client_tcp_sock = accept(tcp_sock, 0, 0)) > 0) )
#else
		if ((msgsock = accept(sock, 0, 0)) > 0)
#endif
		{
			if (msgsock > 0)
				scan_type = SCAN_LOCAL;
#ifdef SOPHIE_NET
			if (client_tcp_sock > 0)
				scan_type = SCAN_NETWORK;
#endif

			sophie_print(1, "accept() set - scan type [%d]", scan_type);

			/*
			   If we got EINTR, it means some signal is caught -
			   on HUP, we'll reload (and ignore this error), and on
			   other signal we'll die anyway (probably :)
			*/

			if ((scan_type == SCAN_LOCAL) && (msgsock == -1 && errno != EINTR))
			{
				sophie_print(0, "error in local accept() - exiting");
			}
			else if (scan_type == SCAN_LOCAL && msgsock == -1) /* this should be EINTR now */
			{
				sophie_print(1, "EINTR set/caught in local accept() loop - some signal received");
				continue;
			}
#ifdef SOPHIE_NET
			else if ((scan_type == SCAN_NETWORK) && (client_tcp_sock == -1 && errno != EINTR))
			{
				sophie_print(0, "error in network accept() - exiting");
			}
			else if (scan_type == SCAN_NETWORK && client_tcp_sock == -1)
			{
				sophie_print(1, "EINTR set/caught in network accept() loop - some signal received");
			}
#endif
			else
			{
				/* Suggested by philipp@corpex.de (Philipp Gaschtz) */
				while( (PROC_COUNT >= MAX_PROC) && (MAX_PROC > 0) )
					usleep(10);

/* ############################# */

			/* Fork */
			sigemptyset(&bset);
			sigaddset(&bset, SIGCHLD);
			sigprocmask(SIG_BLOCK, &bset, NULL);

			cpid = fork();

			if (cpid == -1)
			{
				sigprocmask(SIG_UNBLOCK,&bset,NULL);
				sophie_print(1, "fork() failed! [%s]", strerror(errno));
				close(msgsock);
				goto LOOP;
			}
			else if (cpid > 0)
			{
				close(msgsock);
				PROC_COUNT++;
				sigprocmask(SIG_UNBLOCK,&bset,NULL);
				sophie_print(1, "forked() a child - everything seems ok");
				continue;
			}
			else if (cpid == 0)
			{
				/* child */
				int ret;

				sigprocmask(SIG_UNBLOCK,&bset,NULL);

				/* Print the PROC_COUNT */
				sophie_print(1, "Current PROC_COUNT: [%d]", PROC_COUNT);

				/* We do setuid() after PIDFILE was created (/var/run is owned by root) */
				userInfo = getpwnam(RUNAS_USER);

				/* Check if the userInfo->pw_gid is valid (it should never be NULL, but...) */
				if (userInfo->pw_gid)
				{
					if ((setgid(userInfo->pw_gid)) == -1)
						sophie_print(0, "ERROR: failed to setgid() to [%d]", userInfo->pw_gid);
					else
						sophie_print(1, "NOTE : Successful setgid() to [%d]", userInfo->pw_gid);
				}
				else
				{
					sophie_print(1, "ERROR: userInfo->pwgid is not valid");
				}

				/* Init supplementary group access list*/
				/* Seems like only root can use this? */
				if (getuid() == 0)
				{
					if (initgroups(RUNAS_USER, userInfo->pw_gid) == -1)
					{
						sophie_print(0, "ERROR: initgroups() failed");
						exit(-1);
					}
				}

				if ((setuid(userInfo->pw_uid)) == -1)
					sophie_print(0, "ERROR: failed to setuid() to [%d]", userInfo->pw_uid);
				else
					sophie_print(1, "NOTE : Successful setuid() to [%d]", userInfo->pw_uid);

				/*	bzero((char *)(&sa_exit), sizeof(sa_exit)); */
				sigemptyset(&sa_handle.sa_mask);
				sa_handle.sa_handler = sophie_timeout;

				if ((sig_ret = sigaction(SIGALRM, &sa_handle, NULL)) < 0)
					sophie_print(0, "Error setting handler for SIGALRM");

				/* Clean the buffer */				
				memset(buf, 0, sizeof(buf));
				sophie_print(1, "cleared buf using memset()");

#ifdef SOPHIE_NET
				if (scan_type == SCAN_NETWORK)
				{
					sophie_net_scan(client_tcp_sock);
					shutdown(client_tcp_sock, 2);
					close(client_tcp_sock);
					exit(0);
				}
				else
				{
#endif
					/*
					  We will set the alarm (timeout)
					  If Sophie doesn't receive anything within SOPHIE_TIMEOUT seconds - quit
					*/
					alarm(SOPHIE_TIMEOUT);

					while((ret = sophie_getline(buf, sizeof(buf), msgsock)))
					{
						sophie_print(1, "read %d bytes from socket", ret);
					
						/* Clear the virus name */
						memset(VIR_NAME, 0, sizeof(VIR_NAME));

						/* Clear the ret_error_string */
						memset(ret_error_string, 0, sizeof(ret_error_string));

						/* Remove the newline (if it exists) */
						if (strchr(buf, '\n'))
							*strchr(buf, '\n') = '\0';
					
						sophie_print(1, "[%d] data read [%s]", getpid(), buf);
					
						/* Now, let's decide if we'll scan file or a dir... */
						if ((stat(buf, &reqstat)) == 0)
						{
							if (S_ISDIR(reqstat.st_mode))
							{
								sophos_result = sophie_scandir(buf);
							}
							else if (S_ISREG(reqstat.st_mode))
							{
								sophos_result = sophie_scanfile(buf);
							}
							else
							{
								sophie_print(0, "Error : What was requested is not file or directory - aborting");
								snprintf(ret_error_string, sizeof(ret_error_string)-1, "'%s' not scanned (not a file/directory)", buf);
								sophos_result = -1;
							}
						}
						else
						{
							/* File/dir doesn't exist */
							sophie_print(0, "Error : File or directory doesn't exist (%s)", buf);
							snprintf(ret_error_string, sizeof(ret_error_string)-1, "'%s' does not exist", buf);
							sophos_result = -1;
						}

						/* Clear the response buffer */
						memset(sock_response, 0, sizeof(sock_response));

						/* "Craft" the response (this is so lame) - FIXME */
						if ( (VIR_NAME[0] != '\0') && (REPORT_VIRUSNAME == 1) )
							snprintf(sock_response, sizeof(sock_response)-1, "%d:%s", sophos_result, VIR_NAME);
#ifdef RETURN_ERROR_STRING
						else if ( (ret_error_string[0] != '\0') && (RETURN_ERROR_STRING) )
							snprintf(sock_response, sizeof(sock_response)-1, "%d:%s", sophos_result, ret_error_string);
#endif
						else
							snprintf(sock_response, sizeof(sock_response)-1, "%d", sophos_result);

						sophie_print(1, "response = [%s]", sock_response);
						sophie_print(1, "sending the response [%s] to socket", sock_response);

						/* Complain on errors - do nothing if everything is okay */
						if (write(msgsock, sock_response, strlen(sock_response)) > 0)
							sophie_print(1, "response [%s] sent!", sock_response);
						else
							sophie_print(0, "ERROR: write() to remote end of the socket failed!");

						/*
						   Before, we would setup alarm that affects the length of the whole 'communication' between
						   Sophie and the client. Now, we'll make alarm per file (input from the user), so that client
						   can keep (more or less) persistent connection.
						*/
						alarm(SOPHIE_TIMEOUT);
						sophie_print(1, "Alarm (%d) set", SOPHIE_TIMEOUT);
					}

					sophie_print(1, "i am gone...");
				}

				alarm(0);

				/* End of the child process */
				exit(0);
#ifdef SOPHIE_NET
			}
#endif
		}

		if (scan_type == SCAN_LOCAL)
			close(msgsock);

#ifdef SOPHIE_NET
		if (scan_type == SCAN_NETWORK)
			close(client_tcp_sock);
#endif
	}
	}
}
